from swarmtorch.swarm.model_training.pso import PSO
from swarmtorch.swarm.model_training.gwo import GWO
from swarmtorch.swarm.model_training.woa import WOA
from swarmtorch.swarm.model_training.hho import HHO
from swarmtorch.swarm.model_training.ssa import SSA
from swarmtorch.swarm.model_training.firefly import Firefly, Bat, Dragonfly
from swarmtorch.swarm.model_training.cuckoo import CuckooSearch, Salp, Bee, Fish
from swarmtorch.swarm.model_training.remaining import (
    DFO,
    MBO,
    CSA,
    AOA,
    SOS,
    DVBA,
    ABCO,
    GOA,
    HUS,
    JY,
    SPBO,
    RandomSearch,
    IGWO,
    IWOA,
    ACGWO,
    Memetic,
    Clonalg,
)
from swarmtorch.swarm.model_training.extra import AFSA, HSA, DFO2

__all__ = [
    "PSO",
    "GWO",
    "WOA",
    "HHO",
    "SSA",
    "Firefly",
    "Bat",
    "Dragonfly",
    "CuckooSearch",
    "Salp",
    "Bee",
    "Fish",
    "DFO",
    "MBO",
    "CSA",
    "AOA",
    "SOS",
    "DVBA",
    "ABCO",
    "GOA",
    "HUS",
    "JY",
    "SPBO",
    "RandomSearch",
    "IGWO",
    "IWOA",
    "ACGWO",
    "Memetic",
    "Clonalg",
    "AFSA",
    "HSA",
    "DFO2",
]



from typing import Any

import torch

from swarmtorch.base import SwarmOptimizer


class SSA(SwarmOptimizer):
    """Sparrow Search Algorithm (SSA) optimizer for PyTorch models.

    SSA simulates the foraging and anti-predator behavior of sparrows.
    It uses producers (leaders) and scroungers (followers) with
    discovery and joining behaviors.

    Args:
        params: Model parameters to optimize.
        swarm_size: Number of sparrows in the swarm (default: 30).
        device: Device to run computations on (default: "cpu").

    Example:
        >>> model = torch.nn.Linear(10, 2)
        >>> optimizer = SSA(model.parameters(), swarm_size=30)
        >>> for data, target in dataloader:
        ...     def closure():
        ...         output = model(data)
        ...         return loss_fn(output, target)
        ...     optimizer.zero_grad()
        ...     optimizer.step(closure)
    """

    def __init__(
        self,
        params: Any,
        swarm_size: int = 30,
        device: str = "cpu",
    ) -> None:
        dict(
            swarm_size=swarm_size,
            device=device,
        )
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        """Initialize sparrow positions and fitness."""
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]

        self.positions = torch.rand(
            self.swarm_size,
            param_shape[0],
            device=self.device,
        )

        self.fitness = torch.full(
            (self.swarm_size,),
            float("inf"),
            device=self.device,
        )

        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        """Set model parameters from flattened tensor."""
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                numel = p.numel()
                p.data.copy_(flat_params[idx : idx + numel].reshape(p.shape))
                idx += numel

    def _get_params(self) -> torch.Tensor:
        """Get flattened model parameters."""
        params = []
        for group in self.param_groups:
            for p in group["params"]:
                params.append(p.data.flatten())
        return torch.cat(params)

    def _evaluate_fitness(
        self,
        particles: torch.Tensor,
        closure: Any | None = None,
    ) -> torch.Tensor:
        """Evaluate fitness for each sparrow using the closure."""
        if closure is None:
            raise ValueError("SSA requires a closure function to evaluate fitness")

        fitness = torch.zeros(particles.shape[0], device=self.device)

        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            loss = closure()
            fitness[i] = loss.detach()

        return fitness

    def _update_positions(self) -> None:
        """Update sparrow positions based on SSA equations."""
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return

        fitness = self._evaluate_fitness(self.positions, closure)

        sorted_indices = torch.argsort(fitness)
        worst_indices = sorted_indices[-int(self.swarm_size * 0.2) :]
        best_indices = sorted_indices[: int(self.swarm_size * 0.1)]

        best_idx = sorted_indices[0]
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        pd = 0.8

        for i in range(self.swarm_size):
            if i in best_indices:
                r1 = torch.rand(1, device=self.device).item()
                if r1 < pd:
                    self.positions[i] = (
                        self.positions[i]
                        * torch.exp(
                            -self.iteration_count
                            / (torch.rand(1, device=self.device) * max_iter)
                        ).item()
                        + torch.randn(self.positions.shape[1], device=self.device)
                        * 0.01
                    )
                else:
                    self.positions[i] = self.best_position + torch.randn(
                        self.positions.shape[1], device=self.device
                    ) * torch.rand(1, device=self.device)
            elif i in worst_indices:
                self.positions[i] = (
                    self.best_position.reshape(1, -1)
                    + torch.randn(self.positions.shape[1], device=self.device)
                    * torch.abs(self.positions[i] - self.best_position)
                    * (torch.rand(1, device=self.device).item() - 0.5)
                    * 2
                ).squeeze()
            else:
                self.positions[i] = self.positions[i] + (
                    torch.rand(1, device=self.device).item() - 0.5
                ) * 2 * (self.positions[i] - self.best_position) / (
                    fitness[i] - self.best_fitness + 1e-10
                )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any | None = None) -> Any:
        """Perform one SSA optimization step.

        Args:
            closure: A closure that reevaluates the model and returns the loss.

        Returns:
            The loss value if closure is provided, None otherwise.
        """
        self._current_closure = closure
        return super().step(closure)



from swarmtorch.base.generic_search import GenericSwarmSearch
from swarmtorch.physics.model_training import SA, GSA, FPA


class SASearch(GenericSwarmSearch):
    """SASearch hyperparameter search using SA."""

    def __init__(self, *args, **kwargs):
        super().__init__(SA, *args, **kwargs)


class GSASearch(GenericSwarmSearch):
    """GSASearch hyperparameter search using GSA."""

    def __init__(self, *args, **kwargs):
        super().__init__(GSA, *args, **kwargs)


class FPASearch(GenericSwarmSearch):
    """FPASearch hyperparameter search using FPA."""

    def __init__(self, *args, **kwargs):
        super().__init__(FPA, *args, **kwargs)


__all__ = ["SASearch", "GSASearch", "FPASearch"]



from swarmtorch.base.generic_search import GenericSwarmSearch
from swarmtorch.swarm.model_training import (
    PSO,
    GWO,
    WOA,
    HHO,
    SSA,
    Firefly,
    Bat,
    Dragonfly,
    CuckooSearch,
    Salp,
    Bee,
    Fish,
    DFO,
    MBO,
    CSA,
    AOA,
    SOS,
    DVBA,
    ABCO,
    GOA,
    HUS,
    JY,
    SPBO,
    RandomSearch,
    IGWO,
    IWOA,
    ACGWO,
    Memetic,
    Clonalg,
    AFSA,
    HSA,
    DFO2,
)


class PSOSearch(GenericSwarmSearch):
    """PSOSearch hyperparameter search using PSO."""

    def __init__(self, *args, **kwargs):
        super().__init__(PSO, *args, **kwargs)


class GWOSearch(GenericSwarmSearch):
    """GWOSearch hyperparameter search using GWO."""

    def __init__(self, *args, **kwargs):
        super().__init__(GWO, *args, **kwargs)


class WOASearch(GenericSwarmSearch):
    """WOASearch hyperparameter search using WOA."""

    def __init__(self, *args, **kwargs):
        super().__init__(WOA, *args, **kwargs)


class HHOSearch(GenericSwarmSearch):
    """HHOSearch hyperparameter search using HHO."""

    def __init__(self, *args, **kwargs):
        super().__init__(HHO, *args, **kwargs)


class SSASearch(GenericSwarmSearch):
    """SSASearch hyperparameter search using SSA."""

    def __init__(self, *args, **kwargs):
        super().__init__(SSA, *args, **kwargs)


class FireflySearch(GenericSwarmSearch):
    """FireflySearch hyperparameter search using Firefly."""

    def __init__(self, *args, **kwargs):
        super().__init__(Firefly, *args, **kwargs)


class BatSearch(GenericSwarmSearch):
    """BatSearch hyperparameter search using Bat."""

    def __init__(self, *args, **kwargs):
        super().__init__(Bat, *args, **kwargs)


class DragonflySearch(GenericSwarmSearch):
    """DragonflySearch hyperparameter search using Dragonfly."""

    def __init__(self, *args, **kwargs):
        super().__init__(Dragonfly, *args, **kwargs)


class CuckooSearchHT(GenericSwarmSearch):
    """CuckooSearchHT hyperparameter search using CuckooSearch."""

    def __init__(self, *args, **kwargs):
        super().__init__(CuckooSearch, *args, **kwargs)


class SalpSearch(GenericSwarmSearch):
    """SalpSearch hyperparameter search using Salp."""

    def __init__(self, *args, **kwargs):
        super().__init__(Salp, *args, **kwargs)


class BeeSearch(GenericSwarmSearch):
    """BeeSearch hyperparameter search using Bee."""

    def __init__(self, *args, **kwargs):
        super().__init__(Bee, *args, **kwargs)


class FishSearch(GenericSwarmSearch):
    """FishSearch hyperparameter search using Fish."""

    def __init__(self, *args, **kwargs):
        super().__init__(Fish, *args, **kwargs)


class DFOSearch(GenericSwarmSearch):
    """DFOSearch hyperparameter search using DFO."""

    def __init__(self, *args, **kwargs):
        super().__init__(DFO, *args, **kwargs)


class MBOSearch(GenericSwarmSearch):
    """MBOSearch hyperparameter search using MBO."""

    def __init__(self, *args, **kwargs):
        super().__init__(MBO, *args, **kwargs)


class CSASearch(GenericSwarmSearch):
    """CSASearch hyperparameter search using CSA."""

    def __init__(self, *args, **kwargs):
        super().__init__(CSA, *args, **kwargs)


class AOASearch(GenericSwarmSearch):
    """AOASearch hyperparameter search using AOA."""

    def __init__(self, *args, **kwargs):
        super().__init__(AOA, *args, **kwargs)


class SOSSearch(GenericSwarmSearch):
    """SOSSearch hyperparameter search using SOS."""

    def __init__(self, *args, **kwargs):
        super().__init__(SOS, *args, **kwargs)


class DVBASearch(GenericSwarmSearch):
    """DVBASearch hyperparameter search using DVBA."""

    def __init__(self, *args, **kwargs):
        super().__init__(DVBA, *args, **kwargs)


class ABCOSearch(GenericSwarmSearch):
    """ABCOSearch hyperparameter search using ABCO."""

    def __init__(self, *args, **kwargs):
        super().__init__(ABCO, *args, **kwargs)


class GOASearch(GenericSwarmSearch):
    """GOASearch hyperparameter search using GOA."""

    def __init__(self, *args, **kwargs):
        super().__init__(GOA, *args, **kwargs)


class HUSSearch(GenericSwarmSearch):
    """HUSSearch hyperparameter search using HUS."""

    def __init__(self, *args, **kwargs):
        super().__init__(HUS, *args, **kwargs)


class JYSearch(GenericSwarmSearch):
    """JYSearch hyperparameter search using JY."""

    def __init__(self, *args, **kwargs):
        super().__init__(JY, *args, **kwargs)


class SPBOSearch(GenericSwarmSearch):
    """SPBOSearch hyperparameter search using SPBO."""

    def __init__(self, *args, **kwargs):
        super().__init__(SPBO, *args, **kwargs)


class RandomSearchHT(GenericSwarmSearch):
    """RandomSearchHT hyperparameter search using RandomSearch."""

    def __init__(self, *args, **kwargs):
        super().__init__(RandomSearch, *args, **kwargs)


class IGWOSearch(GenericSwarmSearch):
    """IGWOSearch hyperparameter search using IGWO."""

    def __init__(self, *args, **kwargs):
        super().__init__(IGWO, *args, **kwargs)


class IWOASearch(GenericSwarmSearch):
    """IWOASearch hyperparameter search using IWOA."""

    def __init__(self, *args, **kwargs):
        super().__init__(IWOA, *args, **kwargs)


class ACGWOSearch(GenericSwarmSearch):
    """ACGWOSearch hyperparameter search using ACGWO."""

    def __init__(self, *args, **kwargs):
        super().__init__(ACGWO, *args, **kwargs)


class MemeticSearch(GenericSwarmSearch):
    """MemeticSearch hyperparameter search using Memetic."""

    def __init__(self, *args, **kwargs):
        super().__init__(Memetic, *args, **kwargs)


class ClonalgSearch(GenericSwarmSearch):
    """ClonalgSearch hyperparameter search using Clonalg."""

    def __init__(self, *args, **kwargs):
        super().__init__(Clonalg, *args, **kwargs)


class AFSASearch(GenericSwarmSearch):
    """AFSASearch hyperparameter search using AFSA."""

    def __init__(self, *args, **kwargs):
        super().__init__(AFSA, *args, **kwargs)


class HSASearch(GenericSwarmSearch):
    """HSASearch hyperparameter search using HSA."""

    def __init__(self, *args, **kwargs):
        super().__init__(HSA, *args, **kwargs)


class DFO2Search(GenericSwarmSearch):
    """DFO2Search hyperparameter search using DFO2."""

    def __init__(self, *args, **kwargs):
        super().__init__(DFO2, *args, **kwargs)


__all__ = [
    "PSOSearch",
    "GWOSearch",
    "WOASearch",
    "HHOSearch",
    "SSASearch",
    "FireflySearch",
    "BatSearch",
    "DragonflySearch",
    "CuckooSearchHT",
    "SalpSearch",
    "BeeSearch",
    "FishSearch",
    "DFOSearch",
    "MBOSearch",
    "CSASearch",
    "AOASearch",
    "SOSSearch",
    "DVBASearch",
    "ABCOSearch",
    "GOASearch",
    "HUSSearch",
    "JYSearch",
    "SPBOSearch",
    "RandomSearchHT",
    "IGWOSearch",
    "IWOASearch",
    "ACGWOSearch",
    "MemeticSearch",
    "ClonalgSearch",
    "AFSASearch",
    "HSASearch",
    "DFO2Search",
]



from typing import Any
import torch
import math
from swarmtorch.base import SwarmOptimizer


class SineCosine(SwarmOptimizer):
    """Sine Cosine Algorithm (SCA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("SCA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        a = 2
        r1 = a - self.iteration_count * (a / max_iter)

        for i in range(self.swarm_size):
            for j in range(self.positions.shape[1]):
                r2 = torch.rand(1, device=self.device).item() * 2 * 3.14159
                r3 = torch.rand(1, device=self.device).item() * 2
                r4 = torch.rand(1, device=self.device).item()

                if r4 < 0.5:
                    self.positions[i, j] = self.positions[i, j] + r1 * math.sin(
                        r2
                    ) * torch.abs(r3 * self.best_position[j] - self.positions[i, j])
                else:
                    self.positions[i, j] = self.positions[i, j] + r1 * math.cos(
                        r2
                    ) * torch.abs(r3 * self.best_position[j] - self.positions[i, j])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class MFO(SwarmOptimizer):
    """Moth Flame Algorithm (MFO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.flames = torch.rand(
            self.swarm_size // 3, param_shape[0], device=self.device
        )
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("MFO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return

        all_pos = torch.cat([self.positions, self.flames], dim=0)
        fitness = self._evaluate_fitness(all_pos, closure)

        sorted_idx = torch.argsort(fitness)
        self.best_fitness = fitness[sorted_idx[0]]
        self.best_position = all_pos[sorted_idx[0]].clone()

        self.positions = all_pos[: self.swarm_size].clone()
        self.flames = all_pos[self.swarm_size :].clone()

        max_iter = 1000
        t = (self.iteration_count / max_iter) * 2 - 1

        for i in range(self.swarm_size):
            for j in range(self.positions.shape[1]):
                flame_idx = int(i * self.flames.shape[0] / self.swarm_size)
                distance = torch.abs(self.positions[i, j] - self.flames[flame_idx, j])
                b = 1
                t_val = t * (1 - j / self.positions.shape[1])
                self.positions[i, j] = (
                    distance * math.exp(b * t_val) * math.cos(2 * 3.14159 * t_val)
                    + self.flames[flame_idx, j]
                )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from typing import Any
import torch
from swarmtorch.base import SwarmOptimizer


class DFO(SwarmOptimizer):
    """Dingo Optimization Algorithm (DFO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("DFO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            r = torch.rand_like(self.positions[i])
            self.positions[i] = (
                self.best_position
                + r * (self.positions[i] - self.best_position)
                + torch.randn_like(self.positions[i]) * 0.1
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class MBO(SwarmOptimizer):
    """Marriage in Honey Bees Optimization (MBO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("MBO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            queen_idx = torch.randint(0, self.swarm_size // 3, (1,)).item()
            self.positions[i] = self.positions[queen_idx] + torch.rand_like(
                self.positions[i]
            ) * (self.best_position - self.positions[i])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class CSA(SwarmOptimizer):
    """Crow Search Algorithm (CSA) optimizer for PyTorch models."""

    def __init__(
        self, params: Any, swarm_size: int = 30, ap: float = 0.1, device: str = "cpu"
    ) -> None:
        super().__init__(params, swarm_size=swarm_size, device=device, ap=ap)
        self.ap = ap
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("CSA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            if torch.rand(1, device=self.device).item() > self.ap:
                j = torch.randint(0, self.swarm_size, (1,)).item()
                self.positions[i] = self.positions[i] + 2 * torch.rand_like(
                    self.positions[i]
                ) * (self.positions[j] - self.positions[i])
            else:
                self.positions[i] = torch.rand_like(self.positions[i])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class AOA(SwarmOptimizer):
    """Arithmetic Optimization Algorithm (AOA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("AOA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        mu = 0.5
        t = self.iteration_count / max_iter

        for i in range(self.swarm_size):
            for j in range(self.positions.shape[1]):
                r = torch.rand(1, device=self.device).item()
                if r < 0.5:
                    self.positions[i, j] = (
                        self.best_position[j]
                        / (mu + 1e-10)
                        * torch.exp(
                            torch.tensor(-t, dtype=torch.float32, device=self.device)
                        ).item()
                    )
                else:
                    self.positions[i, j] = (
                        self.best_position[j]
                        * torch.exp(
                            torch.tensor(t, dtype=torch.float32, device=self.device)
                        ).item()
                    )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class SOS(SwarmOptimizer):
    """Symbiotic Organisms Search (SOS) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("SOS requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            j = torch.randint(0, self.swarm_size, (1,)).item()
            while j == i:
                j = torch.randint(0, self.swarm_size, (1,)).item()

            r1 = torch.rand(1, device=self.device).item()
            torch.rand(1, device=self.device).item()

            mutualism = (
                (self.positions[i] + self.positions[j])
                / 2
                * (r1 * self.best_position - self.positions[i])
            )
            self.positions[i] = self.positions[i] + mutualism

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class DVBA(SwarmOptimizer):
    """Dwarf Mongoose Optimization (DVBA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("DVBA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            self.positions[i] = self.best_position + torch.randn_like(
                self.positions[i]
            ) * (1 - self.iteration_count / 1000)

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class ABCO(SwarmOptimizer):
    """Artificial Bee Colony Optimization (ABCO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("ABCO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            partner = torch.randint(0, self.swarm_size, (1,)).item()
            phi = torch.rand(1, device=self.device).item() * 2 - 1
            new_pos = self.positions[i] + phi * (
                self.positions[i] - self.positions[partner]
            )
            self._set_params(new_pos)
            if closure().detach() < fitness[i]:
                self.positions[i] = new_pos

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class GOA(SwarmOptimizer):
    """Grasshopper Optimization Algorithm (GOA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("GOA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        c = 1 - self.iteration_count / max_iter

        for i in range(self.swarm_size):
            s = torch.zeros_like(self.positions[i])
            for j in range(self.swarm_size):
                if i != j:
                    dist = torch.norm(self.positions[i] - self.positions[j])
                    s += (
                        (self.positions[j] - self.positions[i])
                        / (dist + 1e-10)
                        * torch.rand(1, device=self.device).item()
                    )

            self.positions[i] = c * s + self.best_position

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class HUS(SwarmOptimizer):
    """Husband and Wife Optimization (HUS) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("HUS requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            if i < self.swarm_size // 2:
                self.positions[i] = self.best_position + torch.rand_like(
                    self.positions[i]
                ) * (self.best_position - self.positions[i])
            else:
                partner = torch.randint(0, self.swarm_size // 2, (1,)).item()
                self.positions[i] = self.positions[i] + torch.rand_like(
                    self.positions[i]
                ) * (self.positions[partner] - self.positions[i])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class JY(SwarmOptimizer):
    """Jaya Algorithm (JY) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.worst_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)
        self.worst_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("JY requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)

        best_idx = torch.argmin(fitness)
        worst_idx = torch.argmax(fitness)

        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()
        if fitness[worst_idx] > self.worst_fitness:
            self.worst_fitness = fitness[worst_idx]
            self.worst_position = self.positions[worst_idx].clone()

        for i in range(self.swarm_size):
            r1 = torch.rand_like(self.positions[i])
            r2 = torch.rand_like(self.positions[i])
            self.positions[i] = (
                self.positions[i]
                + r1 * (self.best_position - self.positions[i])
                - r2 * (self.worst_position - self.positions[i])
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class SPBO(SwarmOptimizer):
    """Sports Optimization (SPBO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("SPBO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            coach_idx = torch.randint(0, self.swarm_size // 3, (1,)).item()
            self.positions[i] = self.best_position + torch.rand_like(
                self.positions[i]
            ) * (self.positions[coach_idx] - self.positions[i])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class RandomSearch(SwarmOptimizer):
    """Random Search (RandomSearch) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("RandomSearch requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        self.positions = torch.rand_like(self.positions)

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class IGWO(SwarmOptimizer):
    """Improved Grey Wolf Optimizer (IGWO/GWO-Improved) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("IGWO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        sorted_idx = torch.argsort(fitness)

        if fitness[sorted_idx[0]] < self.best_fitness:
            self.best_fitness = fitness[sorted_idx[0]]
            self.best_position = self.positions[sorted_idx[0]].clone()

        max_iter = 1000
        a = 2 - self.iteration_count * (2 / max_iter)
        a2 = 2 - self.iteration_count * (2 / max_iter) * 0.5

        for i in range(self.swarm_size):
            for j in range(self.positions.shape[1]):
                r1 = torch.rand(1, device=self.device).item()
                r2 = torch.rand(1, device=self.device).item()

                A1 = 2 * a * r1 - a
                C1 = 2 * r2
                D1 = torch.abs(C1 * self.best_position[j] - self.positions[i, j])
                X1 = self.best_position[j] - A1 * D1

                r1 = torch.rand(1, device=self.device).item()
                r2 = torch.rand(1, device=self.device).item()

                a2 * (2 * r1 - 1)
                2 * r2
                X2 = self.best_position[j] + torch.randn(1, device=self.device).item()

                self.positions[i, j] = (X1 + X2) / 2

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class IWOA(SwarmOptimizer):
    """Improved Whale Optimization Algorithm (IWOA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("IWOA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        a = 2 - self.iteration_count * (2 / max_iter)
        l = torch.rand(1, device=self.device) * 2 - 1
        p = torch.rand(1, device=self.device).item()

        for i in range(self.swarm_size):
            if p < 0.5:
                r = torch.rand(1, device=self.device).item()
                if abs(a) < 1:
                    D = torch.abs(self.best_position - self.positions[i])
                    self.positions[i] = self.best_position - a * D * r
                else:
                    rand_idx = torch.randint(0, self.swarm_size, (1,)).item()
                    self.positions[i] = (
                        self.positions[rand_idx]
                        + torch.randn_like(self.positions[i]) * 0.1
                    )
            else:
                D = torch.abs(self.best_position - self.positions[i])
                self.positions[i] = (
                    D * torch.exp(l) * torch.cos(2 * 3.14159 * l) + self.best_position
                )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class ACGWO(SwarmOptimizer):
    """Accelerated Chaotic Grey Wolf Optimizer (ACGWO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("ACGWO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        sorted_idx = torch.argsort(fitness)

        if fitness[sorted_idx[0]] < self.best_fitness:
            self.best_fitness = fitness[sorted_idx[0]]
            self.best_position = self.positions[sorted_idx[0]].clone()

        alpha = self.positions[sorted_idx[0]]
        beta = self.positions[sorted_idx[1]] if len(sorted_idx) > 1 else alpha
        delta = self.positions[sorted_idx[2]] if len(sorted_idx) > 2 else alpha

        max_iter = 1000
        a = 2 - self.iteration_count * (2 / max_iter)

        for i in range(self.swarm_size):
            for j in range(self.positions.shape[1]):
                r1 = torch.rand(1, device=self.device).item()
                r2 = torch.rand(1, device=self.device).item()

                A1 = 2 * a * r1 - a
                C1 = 2 * r2
                D1 = torch.abs(C1 * alpha[j] - self.positions[i, j])
                X1 = alpha[j] - A1 * D1

                r1 = torch.rand(1, device=self.device).item()
                r2 = torch.rand(1, device=self.device).item()

                A2 = 2 * a * r1 - a
                C2 = 2 * r2
                D2 = torch.abs(C2 * beta[j] - self.positions[i, j])
                X2 = beta[j] - A2 * D2

                r1 = torch.rand(1, device=self.device).item()
                r2 = torch.rand(1, device=self.device).item()

                A3 = 2 * a * r1 - a
                C3 = 2 * r2
                D3 = torch.abs(C3 * delta[j] - self.positions[i, j])
                X3 = delta[j] - A3 * D3

                chaos = (
                    4
                    * torch.rand(1, device=self.device).item()
                    * (1 - torch.rand(1, device=self.device).item())
                )
                self.positions[i, j] = (X1 + X2 + X3) / 3 + chaos * 0.1

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Memetic(SwarmOptimizer):
    """Memetic Algorithm (Memetic) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Memetic requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _local_search(self, position: torch.Tensor, closure: Any) -> torch.Tensor:
        best_pos = position.clone()
        self._set_params(best_pos)
        best_fit = closure().detach()

        for _ in range(5):
            new_pos = best_pos + torch.randn_like(best_pos) * 0.01
            self._set_params(new_pos)
            new_fit = closure().detach()
            if new_fit < best_fit:
                best_fit = new_fit
                best_pos = new_pos.clone()

        return best_pos

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            parent_idx = torch.randint(0, self.swarm_size, (1,)).item()
            child = (self.positions[i] + self.positions[parent_idx]) / 2
            self.positions[i] = self._local_search(child, closure)

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Clonalg(SwarmOptimizer):
    """Clonal Selection Algorithm (Clonalg) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Clonalg requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        sorted_idx = torch.argsort(fitness)

        if fitness[sorted_idx[0]] < self.best_fitness:
            self.best_fitness = fitness[sorted_idx[0]]
            self.best_position = self.positions[sorted_idx[0]].clone()

        n_select = self.swarm_size // 3
        selected = self.positions[sorted_idx[:n_select]]

        for i in range(n_select):
            clone = selected[i].clone() + torch.randn_like(selected[i]) * 0.1
            self._set_params(clone)
            clone_fit = closure().detach()

            if clone_fit < fitness[sorted_idx[i]]:
                self.positions[sorted_idx[i]] = clone
                fitness[sorted_idx[i]] = clone_fit

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from typing import Any

import torch

from swarmtorch.base import SwarmOptimizer


class HHO(SwarmOptimizer):
    """Harris Hawks Optimization (HHO) optimizer for PyTorch models.

    HHO simulates the hunting behavior of Harris' hawks. The algorithm uses
    phases of exploration (perching) and exploitation (diving) with
    random strategies based on prey energy.

    Args:
        params: Model parameters to optimize.
        swarm_size: Number of hawks in the swarm (default: 30).
        device: Device to run computations on (default: "cpu").

    Example:
        >>> model = torch.nn.Linear(10, 2)
        >>> optimizer = HHO(model.parameters(), swarm_size=30)
        >>> for data, target in dataloader:
        ...     def closure():
        ...         output = model(data)
        ...         return loss_fn(output, target)
        ...     optimizer.zero_grad()
        ...     optimizer.step(closure)
    """

    def __init__(
        self,
        params: Any,
        swarm_size: int = 30,
        device: str = "cpu",
    ) -> None:
        dict(
            swarm_size=swarm_size,
            device=device,
        )
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        """Initialize hawk positions and fitness."""
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]

        self.positions = torch.rand(
            self.swarm_size,
            param_shape[0],
            device=self.device,
        )

        self.fitness = torch.full(
            (self.swarm_size,),
            float("inf"),
            device=self.device,
        )

        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        """Set model parameters from flattened tensor."""
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                numel = p.numel()
                p.data.copy_(flat_params[idx : idx + numel].reshape(p.shape))
                idx += numel

    def _get_params(self) -> torch.Tensor:
        """Get flattened model parameters."""
        params = []
        for group in self.param_groups:
            for p in group["params"]:
                params.append(p.data.flatten())
        return torch.cat(params)

    def _evaluate_fitness(
        self,
        particles: torch.Tensor,
        closure: Any | None = None,
    ) -> torch.Tensor:
        """Evaluate fitness for each hawk using the closure."""
        if closure is None:
            raise ValueError("HHO requires a closure function to evaluate fitness")

        fitness = torch.zeros(particles.shape[0], device=self.device)

        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            loss = closure()
            fitness[i] = loss.detach()

        return fitness

    def _update_positions(self) -> None:
        """Update hawk positions based on HHO equations."""
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return

        fitness = self._evaluate_fitness(self.positions, closure)

        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        e = 2 * (1 - self.iteration_count / max_iter)

        for i in range(self.swarm_size):
            torch.rand(1, device=self.device).item()

            if e >= 1:
                q = torch.rand(1, device=self.device).item()
                if q >= 0.5:
                    random_hawk = torch.randint(0, self.swarm_size, (1,)).item()
                    self.positions[i] = self.positions[random_hawk] + torch.rand(
                        self.positions.shape[1], device=self.device
                    ) * 2 * (
                        torch.rand(self.positions.shape[1], device=self.device) - 0.5
                    )
                else:
                    self.positions[i] = self.best_position + torch.rand(
                        self.positions.shape[1], device=self.device
                    ) * 2 * (
                        torch.rand(self.positions.shape[1], device=self.device) - 0.5
                    )
            else:
                if e >= 0.5:
                    jump_strength = 2 * (1 - self.iteration_count / max_iter)
                    rabbit_mean = torch.mean(self.positions, dim=0)
                    self.positions[i] = (
                        self.best_position
                        - self.positions[i]
                        + (
                            rabbit_mean
                            + jump_strength
                            * 2
                            * (
                                torch.rand(self.positions.shape[1], device=self.device)
                                - 0.5
                            )
                        )
                        * (
                            torch.rand(self.positions.shape[1], device=self.device)
                            - 0.5
                        )
                    )
                else:
                    rabbit_mean = torch.mean(self.positions, dim=0)
                    self.positions[i] = self.best_position - e * torch.abs(
                        rabbit_mean
                        - 2
                        * torch.rand(self.positions.shape[1], device=self.device)
                        * self.best_position
                    )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any | None = None) -> Any:
        """Perform one HHO optimization step.

        Args:
            closure: A closure that reevaluates the model and returns the loss.

        Returns:
            The loss value if closure is provided, None otherwise.
        """
        self._current_closure = closure
        return super().step(closure)



from typing import Any
import torch
from swarmtorch.base import SwarmOptimizer


class EHO(SwarmOptimizer):
    """Elephant Herding Optimization (EHO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("EHO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        cl = 0.5
        for i in range(1, self.swarm_size):
            self.positions[i] = (
                cl * self.positions[i] + (1 - cl) * self.positions[best_idx]
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class ChickenSwarm(SwarmOptimizer):
    """Chicken Swarm Optimization (CSO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("CSO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        num_roosters = max(1, self.swarm_size // 10)
        for i in range(self.swarm_size):
            if i < num_roosters:
                self.positions[i] = (
                    self.positions[i] + torch.randn_like(self.positions[i]) * 0.5
                )
            else:
                mother_idx = torch.randint(0, num_roosters, (1,)).item()
                self.positions[i] = self.positions[i] + torch.rand_like(
                    self.positions[i]
                ) * (self.positions[mother_idx] - self.positions[i])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class SMA(SwarmOptimizer):
    """Slime Mold Algorithm (SMA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("SMA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        sorted_idx = torch.argsort(fitness)
        if fitness[sorted_idx[0]] < self.best_fitness:
            self.best_fitness = fitness[sorted_idx[0]]
            self.best_position = self.positions[sorted_idx[0]].clone()

        w = torch.exp(
            torch.arange(self.swarm_size, device=self.device).float()
            / (-self.swarm_size)
        )

        for i in range(self.swarm_size):
            p = torch.tanh(torch.abs(fitness[i]))
            vb = torch.rand(self.positions.shape[1], device=self.device) * 2 - 1
            vc = torch.rand(self.positions.shape[1], device=self.device) * 2 - 1

            idx1 = torch.randint(0, self.swarm_size, (1,)).item()
            idx2 = torch.randint(0, self.swarm_size, (1,)).item()

            self.positions[i] = (
                self.best_position
                + vb * p * (self.positions[idx1] - self.positions[idx2])
                + vc * w[i] * (self.best_position - self.positions[i])
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from collections.abc import Callable
from typing import Any

import torch
import torch.nn as nn

from swarmtorch.base.hyperparam_search import HyperparameterSearch
from swarmtorch.base.swarm_optimizer import SwarmOptimizer


class _DummyModel(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        # Initializing close to 0.5 to keep it in the center of [0, 1] bounds initially
        self.params = nn.Parameter(torch.rand(dim))


class GenericSwarmSearch(HyperparameterSearch):
    """Generic adapter to use any SwarmOptimizer for hyperparameter tuning.

    This class wraps a standard SwarmOptimizer (like PSO, GWO) and allows it
    to optimize hyperparameters that are continuous, discrete, or categorical
    by providing a translation layer (a dummy parameter space in [0, 1]).
    """

    def __init__(
        self,
        optimizer_class: type[SwarmOptimizer],
        model_fn: Callable[[dict], torch.nn.Module],
        param_space: dict[str, tuple[Any, ...] | list[Any]],
        train_fn: Callable[[torch.nn.Module, dict], float],
        iterations: int = 50,
        swarm_size: int = 30,
        device: str = "cpu",
        verbose: bool = True,
        **optimizer_kwargs: Any,
    ) -> None:
        super().__init__(
            model_fn=model_fn,
            param_space=param_space,
            train_fn=train_fn,
            iterations=iterations,
            swarm_size=swarm_size,
            device=device,
            verbose=verbose,
        )
        self.optimizer_class = optimizer_class
        self.optimizer_kwargs = optimizer_kwargs

    def search(self) -> dict[str, Any]:
        """Run the generic hyperparameter search using the wrapped optimizer."""
        dim = len(self.param_space)
        dummy_model = _DummyModel(dim).to(self.device)

        import inspect

        kwargs = self.optimizer_kwargs.copy()
        sig = inspect.signature(self.optimizer_class.__init__)
        params = sig.parameters
        has_varkw = any(
            p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()
        )

        if "device" in params or has_varkw:
            kwargs["device"] = str(self.device)

        if "swarm_size" in params or has_varkw:
            kwargs["swarm_size"] = self.swarm_size
        elif "population_size" in params:
            kwargs["population_size"] = self.swarm_size

        # Initialize the raw metaheuristic optimizer
        optimizer = self.optimizer_class(dummy_model.parameters(), **kwargs)

        # We need to capture the global best ourselves since optimizers track it differently
        global_best_encoded = torch.zeros(dim, device=self.device)
        global_best_score = float("inf")

        for iteration in range(self.iterations):

            def closure() -> torch.Tensor:
                nonlocal global_best_score, global_best_encoded

                # Enforce [0, 1] bounds directly on the dummy parameters
                with torch.no_grad():
                    dummy_model.params.clamp_(0.0, 1.0)
                    encoded = dummy_model.params.clone().detach()

                try:
                    score = self._evaluate(encoded)
                except Exception:
                    score = float("inf")

                # Track best score explicitly, as different optimizers might lose it or track it differently
                if score < global_best_score:
                    global_best_score = score
                    global_best_encoded = encoded.clone()

                return torch.tensor(score, device=self.device, requires_grad=False)

            # Perform one optimization step
            optimizer.step(closure)

            # Help the optimizer stay within bounds by clamping its internal representations if possible
            if hasattr(optimizer, "positions"):
                with torch.no_grad():
                    optimizer.positions.clamp_(0.0, 1.0)
            elif hasattr(optimizer, "population"):
                with torch.no_grad():
                    optimizer.population.clamp_(0.0, 1.0)

            # Log current best
            self._log(
                iteration, global_best_score, self._decode_params(global_best_encoded)
            )

        self.best_params = self._decode_params(global_best_encoded)
        self.best_score = global_best_score
        return self.best_params



from typing import Any
import torch
from swarmtorch.base import SwarmOptimizer


class Firefly(SwarmOptimizer):
    """Firefly Algorithm (FA) optimizer for PyTorch models."""

    def __init__(
        self,
        params: Any,
        swarm_size: int = 30,
        alpha: float = 0.5,
        beta0: float = 1.0,
        gamma: float = 1.0,
        device: str = "cpu",
    ) -> None:
        super().__init__(
            params,
            swarm_size=swarm_size,
            device=device,
            alpha=alpha,
            beta0=beta0,
            gamma=gamma,
        )
        self.alpha = alpha
        self.beta0 = beta0
        self.gamma = gamma
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("FA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            for j in range(self.swarm_size):
                if fitness[j] < fitness[i]:
                    dist = torch.norm(self.positions[i] - self.positions[j])
                    beta = self.beta0 * torch.exp(-self.gamma * dist**2)
                    self.positions[i] = (
                        self.positions[i]
                        + beta * (self.positions[j] - self.positions[i])
                        + self.alpha * (torch.rand_like(self.positions[i]) - 0.5)
                    )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Bat(SwarmOptimizer):
    """Bat Algorithm (BA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.velocities = torch.zeros_like(self.positions)
        self.frequencies = torch.zeros(self.swarm_size, device=self.device)
        self.loudnesses = torch.full((self.swarm_size,), 0.5, device=self.device)
        self.pulse_rates = torch.full((self.swarm_size,), 0.5, device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("BA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            self.frequencies[i] = 0 + 2 * torch.rand(1, device=self.device).item()
            self.velocities[i] += (
                self.positions[i] - self.best_position
            ) * self.frequencies[i]
            self.positions[i] = self.positions[i] + self.velocities[i]

            if torch.rand(1, device=self.device).item() > self.pulse_rates[i]:
                self.positions[i] = self.best_position + 0.001 * torch.randn_like(
                    self.positions[i]
                )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Dragonfly(SwarmOptimizer):
    """Dragonfly Algorithm (DA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("DA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            neighbor_idx = torch.randint(0, self.swarm_size, (3,)).tolist()
            separation = -torch.sum(self.positions[i] - self.positions[neighbor_idx[0]])
            alignment = torch.mean(self.positions[neighbor_idx[1]]) - self.positions[i]
            cohesion = torch.mean(self.positions[neighbor_idx[2]]) - self.positions[i]

            self.positions[i] += (separation + alignment + cohesion) * 0.1 + (
                self.best_position - self.positions[i]
            ) * 0.01

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from swarmtorch.base.generic_search import GenericSwarmSearch
from swarmtorch.bio_inspired.model_training import ALO, BBO, MVO, SineCosine, MFO


class ALOSearch(GenericSwarmSearch):
    """ALOSearch hyperparameter search using ALO."""

    def __init__(self, *args, **kwargs):
        super().__init__(ALO, *args, **kwargs)


class BBOSearch(GenericSwarmSearch):
    """BBOSearch hyperparameter search using BBO."""

    def __init__(self, *args, **kwargs):
        super().__init__(BBO, *args, **kwargs)


class MVOSearch(GenericSwarmSearch):
    """MVOSearch hyperparameter search using MVO."""

    def __init__(self, *args, **kwargs):
        super().__init__(MVO, *args, **kwargs)


class SineCosineSearch(GenericSwarmSearch):
    """SineCosineSearch hyperparameter search using SineCosine."""

    def __init__(self, *args, **kwargs):
        super().__init__(SineCosine, *args, **kwargs)


class MFOSearch(GenericSwarmSearch):
    """MFOSearch hyperparameter search using MFO."""

    def __init__(self, *args, **kwargs):
        super().__init__(MFO, *args, **kwargs)


__all__ = ["ALOSearch", "BBOSearch", "MVOSearch", "SineCosineSearch", "MFOSearch"]



from .hyperparam_search import HyperparameterSearch
from .swarm_optimizer import SwarmOptimizer
from .generic_search import GenericSwarmSearch

__all__ = ["SwarmOptimizer", "HyperparameterSearch", "GenericSwarmSearch"]



from typing import Any
import torch
import math
from swarmtorch.base import SwarmOptimizer


class CuckooSearch(SwarmOptimizer):
    """Cuckoo Search (CS) optimizer for PyTorch models."""

    def __init__(
        self, params: Any, swarm_size: int = 30, pa: float = 0.25, device: str = "cpu"
    ) -> None:
        super().__init__(params, swarm_size=swarm_size, device=device, pa=pa)
        self.pa = pa
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("CS requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _levy_flight(self, dim: int) -> torch.Tensor:
        beta = 1.5
        sigma = (
            math.gamma(1 + beta)
            * math.sin(math.pi * beta / 2)
            / (math.gamma((1 + beta) / 2) * beta * 2 ** ((beta - 1) / 2))
        ) ** (1 / beta)
        u = torch.randn(dim, device=self.device) * sigma
        v = torch.randn(dim, device=self.device)
        step = u / (v.abs() ** (1 / beta))
        return step

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            if torch.rand(1, device=self.device).item() > self.pa:
                self.positions[i] = self.positions[i] + 0.01 * self._levy_flight(
                    self.positions.shape[1]
                ) * (self.positions[i] - self.best_position)

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Salp(SwarmOptimizer):
    """Salp Chain Algorithm (Salp) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Salp requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        c1 = 2 * math.exp(-((4 * self.iteration_count / 1000) ** 2))

        for i in range(self.swarm_size):
            if i == 0:
                self.positions[i] = self.best_position + c1 * (
                    torch.rand(self.positions.shape[1], device=self.device) * 2 - 1
                )
            else:
                self.positions[i] = (
                    self.positions[i] + self.positions[i - 1]
                ) / 2 + c1 * (
                    torch.rand(self.positions.shape[1], device=self.device) * 2 - 1
                )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Bee(SwarmOptimizer):
    """Artificial Bee Colony (ABC/Bee) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.trials = torch.zeros(self.swarm_size, device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Bee requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            j = torch.randint(0, self.positions.shape[1], (1,)).item()
            k = torch.randint(0, self.swarm_size, (1,)).item()
            while k == i:
                k = torch.randint(0, self.swarm_size, (1,)).item()

            new_pos = self.positions[i].clone()
            new_pos[j] = self.positions[i, j] + (
                torch.rand(1, device=self.device).item() * 2 - 1
            ) * (self.positions[i, j] - self.positions[k, j])

            self._set_params(new_pos)
            new_fitness = closure().detach()

            if new_fitness < fitness[i]:
                self.positions[i] = new_pos
                self.trials[i] = 0
            else:
                self.trials[i] += 1

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Fish(SwarmOptimizer):
    """Artificial Fish Swarm (AFS/Fish) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Fish requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            center = torch.mean(self.positions, dim=0)
            self._set_params(center)
            center_fitness = closure().detach()

            if center_fitness < fitness[i]:
                self.positions[i] = center + torch.rand_like(
                    self.positions[i]
                ) * 0.5 * (center - self.positions[i])
            else:
                partner_idx = torch.randint(0, self.swarm_size, (1,)).item()
                self.positions[i] = self.positions[i] + torch.rand_like(
                    self.positions[i]
                ) * (self.positions[partner_idx] - self.positions[i])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from torch import nn
from swarmtorch.base.generic_search import GenericSwarmSearch
from swarmtorch.swarm.model_training.pso import PSO


def test_generic_search():
    def build_model(params):
        return nn.Linear(params["in_features"], 2)

    param_space = {"in_features": [10, 20, 30], "lr": (0.001, 0.1)}

    def train_fn(model, params):
        return abs(params["lr"] - 0.05) + abs(params["in_features"] - 20)

    search = GenericSwarmSearch(
        optimizer_class=PSO,
        model_fn=build_model,
        param_space=param_space,
        train_fn=train_fn,
        iterations=5,
        swarm_size=5,
        device="cpu",
        verbose=False,
    )

    best_params = search.search()
    assert "lr" in best_params
    assert "in_features" in best_params



from typing import Any
import torch
from swarmtorch.base import SwarmOptimizer


class GSA(SwarmOptimizer):
    """Gravitational Search Algorithm (GSA) optimizer for PyTorch models."""

    def __init__(
        self,
        params: Any,
        swarm_size: int = 30,
        g0: float = 100.0,
        device: str = "cpu",
    ) -> None:
        super().__init__(params, swarm_size=swarm_size, device=device, g0=g0)
        self.g0 = g0
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]

        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.velocities = torch.zeros_like(self.positions)
        self.fitness = torch.full((self.swarm_size,), float("inf"), device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                numel = p.numel()
                p.data.copy_(flat_params[idx : idx + numel].reshape(p.shape))
                idx += numel

    def _get_params(self) -> torch.Tensor:
        params = []
        for group in self.param_groups:
            for p in group["params"]:
                params.append(p.data.flatten())
        return torch.cat(params)

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("GSA requires a closure function")

        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            loss = closure()
            fitness[i] = loss.detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return

        fitness = self._evaluate_fitness(self.positions, closure)

        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        g = self.g0 / (self.iteration_count + 1)

        masses = (fitness - torch.max(fitness)) / (
            torch.min(fitness) - torch.max(fitness) + 1e-10
        )
        masses = torch.exp(masses)
        masses = masses / (torch.sum(masses) + 1e-10)

        for i in range(self.swarm_size):
            force = torch.zeros_like(self.positions[i])
            for j in range(self.swarm_size):
                if i != j:
                    dist = torch.norm(self.positions[j] - self.positions[i]) + 1e-10
                    force += (
                        torch.rand_like(self.positions[i])
                        * g
                        * masses[j]
                        * (self.positions[j] - self.positions[i])
                        / dist
                    )

            self.velocities[i] = (
                torch.rand_like(self.velocities[i]) * self.velocities[i] + force
            )
            self.positions[i] = self.positions[i] + self.velocities[i]

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from swarmtorch.swarm.model_training import (
    PSO,
    GWO,
    WOA,
    HHO,
    SSA,
    Firefly,
    Bat,
    Dragonfly,
)

__all__ = ["PSO", "GWO", "WOA", "HHO", "SSA", "Firefly", "Bat", "Dragonfly"]



from typing import Any
import torch
from swarmtorch.base import SwarmOptimizer


class CEM(SwarmOptimizer):
    """Cross-Entropy Method (CEM) optimizer for PyTorch models."""

    def __init__(
        self, params: Any, population_size: int = 30, device: str = "cpu"
    ) -> None:
        super().__init__(params, swarm_size=population_size, device=device)
        self.population_size = population_size
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.population_size = self.defaults["swarm_size"]
        self.positions = torch.zeros(
            self.population_size, param_shape[0], device=self.device
        )
        self.samples = torch.zeros(
            self.population_size, param_shape[0], device=self.device
        )
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)
        self.mean = torch.zeros(param_shape[0], device=self.device)
        self.std = torch.ones(param_shape[0], device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, samples: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("CEM requires a closure function")
        fitness = torch.zeros(samples.shape[0], device=self.device)
        for i in range(samples.shape[0]):
            self._set_params(samples[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return

        self.samples = self.mean + self.std * torch.randn(
            self.population_size, self.mean.shape[0], device=self.device
        )
        fitness = self._evaluate_fitness(self.samples, closure)

        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.samples[best_idx].clone()

        n_elite = self.population_size // 5
        elite_idx = torch.argsort(fitness)[:n_elite]
        elite_samples = self.samples[elite_idx]

        self.mean = torch.mean(elite_samples, dim=0)
        self.std = torch.std(elite_samples, dim=0) + 0.01

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class PFA(SwarmOptimizer):
    """Peacock Algorithm (PFA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("PFA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            self.positions[i] = self.best_position + torch.randn_like(
                self.positions[i]
            ) * (1 - self.iteration_count / 1000)

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class ARS(SwarmOptimizer):
    """Archimedes Optimization Algorithm (ARS) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.velocities = torch.zeros_like(self.positions)
        self.density = torch.rand(self.swarm_size, device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("ARS requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)

        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        min_fit = torch.min(fitness)
        max_fit = torch.max(fitness)
        self.density = (fitness - min_fit) / (max_fit - min_fit + 1e-10)

        for i in range(self.swarm_size):
            force = torch.zeros_like(self.positions[i])
            for j in range(self.swarm_size):
                if i != j:
                    force += (self.positions[j] - self.positions[i]) * (
                        self.density[j] - self.density[i]
                    )

            self.velocities[i] = self.velocities[i] * 0.5 + force * 0.1
            self.positions[i] = self.positions[i] + self.velocities[i]

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class FDA(SwarmOptimizer):
    """Forest Defense Algorithm (FDA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("FDA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            center = torch.mean(self.positions, dim=0)
            self.positions[i] = center + torch.randn_like(self.positions[i]) * (
                1 - self.iteration_count / 1000
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class CA(SwarmOptimizer):
    """Cultural Algorithm (CA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)
        self.belief_space = torch.zeros(param_shape[0], device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("CA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()
            self.belief_space = self.best_position.clone()

        for i in range(self.swarm_size):
            self.positions[i] = (
                self.belief_space + torch.randn_like(self.positions[i]) * 0.5
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class PBIL(SwarmOptimizer):
    """Population-Based Incremental Learning (PBIL) optimizer for PyTorch models."""

    def __init__(
        self, params: Any, population_size: int = 30, device: str = "cpu"
    ) -> None:
        super().__init__(params, swarm_size=population_size, device=device)
        self.population_size = population_size
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.population_size = self.defaults["swarm_size"]
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)
        self.probability = torch.ones(param_shape[0], device=self.device) * 0.5

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, samples: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("PBIL requires a closure function")
        fitness = torch.zeros(samples.shape[0], device=self.device)
        for i in range(samples.shape[0]):
            self._set_params(samples[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return

        samples = (
            torch.rand(
                self.population_size, self.probability.shape[0], device=self.device
            )
            < self.probability
        ).float()
        fitness = self._evaluate_fitness(samples, closure)

        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = samples[best_idx].clone()

        self.probability = 0.1 * self.best_position + 0.9 * self.probability

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)




lines = (
    open(r"..\test_tuning_results.txt", "r", encoding="utf-16le").read().splitlines()
)

failed = []
success = 0
for line in lines:
    if "SUCCESS" in line:
        success += 1
    if line.startswith("- "):
        failed.append(line)

with open(r"build_report.txt", "w", encoding="utf-8") as f:
    f.write(f"Success Count: {success}\n")
    f.write("Failed Tuners:\n")
    for failure in failed:
        f.write(failure + "\n")



from typing import Any
import torch
from swarmtorch.base import SwarmOptimizer


class SA(SwarmOptimizer):
    """Simulated Annealing (SA) optimizer for PyTorch models.

    SA is a probabilistic optimization algorithm inspired by the annealing
    process in metallurgy.
    """

    def __init__(
        self,
        params: Any,
        initial_temp: float = 1000.0,
        cooling_rate: float = 0.95,
        device: str = "cpu",
    ) -> None:
        super().__init__(
            params,
            swarm_size=1,
            device=device,
            initial_temp=initial_temp,
            cooling_rate=cooling_rate,
        )
        self.initial_temp = initial_temp
        self.cooling_rate = cooling_rate
        self.temperature = initial_temp
        self.current_params: torch.Tensor | None = None
        self.current_loss: float | None = None

    def _init_swarm(self) -> None:
        self.current_params = self._get_params().clone()
        self.temperature = self.initial_temp

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                numel = p.numel()
                p.data.copy_(flat_params[idx : idx + numel].reshape(p.shape))
                idx += numel

    def _get_params(self) -> torch.Tensor:
        params = []
        for group in self.param_groups:
            for p in group["params"]:
                params.append(p.data.flatten())
        return torch.cat(params)

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None or self.current_params is None:
            return

        self._set_params(self.current_params)
        current_loss_val = closure().item()

        new_params = (
            self.current_params
            + torch.randn_like(self.current_params)
            * self.temperature
            / self.initial_temp
        )
        self._set_params(new_params)
        new_loss_val = closure().item()

        delta = new_loss_val - current_loss_val

        if (
            delta < 0
            or torch.rand(1).item()
            < torch.exp(torch.tensor(-delta / self.temperature)).item()
        ):
            self.current_params = new_params.clone()
            self.current_loss = new_loss_val
        else:
            self._set_params(self.current_params)

        self.temperature *= self.cooling_rate

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        if not self._swarm_initialized:
            self._init_swarm()
            self._swarm_initialized = True

        loss = None
        if closure is not None:
            loss = closure()

        self._update_positions()

        return loss



from collections.abc import Callable
from typing import Any

import torch


class HyperparameterSearch:
    """Base class for hyperparameter search using metaheuristic algorithms.

    Provides a common interface for using swarm/evolutionary algorithms
    to search for optimal hyperparameters.
    """

    def __init__(
        self,
        model_fn: Callable[[dict], torch.nn.Module],
        param_space: dict[str, tuple[Any, ...]],
        train_fn: Callable[[torch.nn.Module, dict], float],
        iterations: int = 50,
        swarm_size: int = 30,
        device: str = "cpu",
        verbose: bool = True,
    ) -> None:
        """Initialize hyperparameter search.

        Args:
            model_fn: Function that creates a model given hyperparameters.
                Signature: (hyperparams: dict) -> torch.nn.Module
            param_space: Dictionary defining the search space.
                Format: {'param_name': (min, max)} for floats,
                       {'param_name': [val1, val2, ...]} for discrete
            train_fn: Function to train model and return score.
                Signature: (model: torch.nn.Module, hyperparams: dict) -> float
            iterations: Number of search iterations.
            swarm_size: Number of particles/agents in the swarm.
            device: Device to run computations on.
            verbose: Whether to print progress.
        """
        self.model_fn = model_fn
        self.param_space = param_space
        self.train_fn = train_fn
        self.iterations = iterations
        self.swarm_size = swarm_size
        self.device = torch.device(device)
        self.verbose = verbose
        self.best_params: dict | None = None
        self.best_score: float | None = None

    def _encode_params(self, params: dict) -> torch.Tensor:
        """Encode hyperparameters to a tensor."""
        encoded = []
        for key, value in self.param_space.items():
            if isinstance(value, list):
                idx = value.index(params[key]) if params[key] in value else 0
                encoded.append(idx / len(value))
            else:
                normalized = (params[key] - value[0]) / (value[1] - value[0])
                encoded.append(normalized)
        return torch.tensor(encoded, dtype=torch.float32, device=self.device)

    def _decode_params(self, encoded: torch.Tensor) -> dict:
        """Decode tensor back to hyperparameters."""
        decoded = {}
        for i, (key, value) in enumerate(self.param_space.items()):
            if isinstance(value, list):
                idx = int(encoded[i].item() * (len(value) - 1))
                decoded[key] = value[min(idx, len(value) - 1)]
            else:
                decoded[key] = value[0] + encoded[i].item() * (value[1] - value[0])
        return decoded

    def _evaluate(self, encoded: torch.Tensor) -> float:
        """Evaluate a single set of hyperparameters."""
        params = self._decode_params(encoded)
        model = self.model_fn(params)
        model = model.to(self.device)
        score = self.train_fn(model, params)
        return score

    def search(self) -> dict[str, Any]:
        """Run the hyperparameter search.

        Returns:
            Dictionary containing best hyperparameters found.
        """
        raise NotImplementedError("Subclasses must implement search")

    def _log(self, iteration: int, score: float, params: dict) -> None:
        """Log progress if verbose is enabled."""
        if self.verbose:
            print(f"Iteration {iteration}: score={score:.4f}, params={params}")



from typing import Any
import torch
from swarmtorch.base import SwarmOptimizer


class CatSwarm(SwarmOptimizer):
    """Cat Swarm Optimization (CSO/CatSwarm) optimizer for PyTorch models."""

    def __init__(
        self, params: Any, swarm_size: int = 30, smp: int = 5, device: str = "cpu"
    ) -> None:
        super().__init__(params, swarm_size=swarm_size, device=device, smp=smp)
        self.smp = smp
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.velocities = torch.zeros_like(self.positions)
        self.flags = torch.randint(0, 2, (self.swarm_size,), device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("CatSwarm requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            if self.flags[i] == 0:
                dim = self.positions.shape[1]
                idx = torch.randint(0, dim, (self.smp,)).tolist()
                candidate = self.positions[i].clone()
                for j in idx:
                    candidate[j] = self.best_position[j]
                self._set_params(candidate)
                if closure().detach() < fitness[i]:
                    self.positions[i] = candidate
            else:
                self.positions[i] = self.positions[i] + self.velocities[i]

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Cockroach(SwarmOptimizer):
    """Cockroach Swarm Optimization (CSO/Cockroach) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Cockroach requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            j = torch.randint(0, self.swarm_size, (1,)).item()
            k = torch.randint(0, self.swarm_size, (1,)).item()
            self.positions[i] = (
                self.positions[i]
                + 0.3 * (self.best_position - self.positions[i])
                + 0.1 * (self.positions[j] - self.positions[k])
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Coati(SwarmOptimizer):
    """Coati Optimization Algorithm (COA/Coati) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Coati requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        t = self.iteration_count / max_iter

        for i in range(self.swarm_size):
            if t < 0.5:
                j = torch.randint(0, self.swarm_size, (1,)).item()
                self.positions[i] = self.positions[i] + torch.rand_like(
                    self.positions[i]
                ) * (self.positions[j] - self.positions[i])
            else:
                self.positions[i] = (
                    self.best_position
                    + (torch.rand(self.positions.shape[1], device=self.device) * 2 - 1)
                    * t
                )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class GorillaTroopsOptimizer(SwarmOptimizer):
    """Gorilla Troops Optimization (GTO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("Gorilla requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        max_iter = 1000
        p = 0.03
        g = 0.5
        self.iteration_count / max_iter

        for i in range(self.swarm_size):
            if torch.rand(1, device=self.device).item() < p:
                self.positions[i] = g * torch.rand(
                    self.positions.shape[1], device=self.device
                )
            else:
                c1 = torch.rand(1, device=self.device).item()
                c2 = torch.rand(1, device=self.device).item()
                if c1 > 0.5:
                    self.positions[i] = self.best_position + c2 * (
                        torch.rand(self.positions.shape[1], device=self.device) * 2 - 1
                    )
                else:
                    self.positions[i] = self.positions[i] - c2 * (
                        self.best_position - self.positions[i]
                    )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class Gorilla(GorillaTroopsOptimizer):
    """Gorilla Troops Optimization (GTO/Gorilla) optimizer for PyTorch models."""

    pass


class JSO(SwarmOptimizer):
    """Jaya Optimization (JSO) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.worst_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)
        self.worst_fitness = torch.tensor(-float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("JSO requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)

        best_idx = torch.argmin(fitness)
        worst_idx = torch.argmax(fitness)

        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()
        if fitness[worst_idx] > self.worst_fitness:
            self.worst_fitness = fitness[worst_idx]
            self.worst_position = self.positions[worst_idx].clone()

        for i in range(self.swarm_size):
            r1 = torch.rand_like(self.positions[i])
            r2 = torch.rand_like(self.positions[i])
            self.positions[i] = (
                self.positions[i]
                + r1 * (self.best_position - self.positions[i])
                - r2 * (self.worst_position - self.positions[i])
            )

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)


class KHA(SwarmOptimizer):
    """Krill Herd Algorithm (KHA) optimizer for PyTorch models."""

    def __init__(self, params: Any, swarm_size: int = 30, device: str = "cpu") -> None:
        super().__init__(params, swarm_size=swarm_size, device=device)
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.velocities = torch.zeros_like(self.positions)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                p.data.copy_(flat_params[idx : idx + p.numel()].reshape(p.shape))
                idx += p.numel()

    def _get_params(self) -> torch.Tensor:
        return torch.cat(
            [p.data.flatten() for group in self.param_groups for p in group["params"]]
        )

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("KHA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            fitness[i] = closure().detach()
        return fitness

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return
        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            center = torch.mean(self.positions, dim=0)
            target_idx = torch.randint(0, self.swarm_size, (1,)).item()

            self.velocities[i] = (
                0.1 * self.velocities[i]
                + 0.5
                * torch.rand_like(self.velocities[i])
                * (self.best_position - self.positions[i])
                + 0.5
                * torch.rand_like(self.velocities[i])
                * (center - self.positions[i])
                + 0.5
                * torch.rand_like(self.velocities[i])
                * (self.positions[target_idx] - self.positions[i])
            )

            self.positions[i] = self.positions[i] + self.velocities[i]

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from swarmtorch.evolutionary.model_training.de import DE
from swarmtorch.evolutionary.model_training.ga import GA
from swarmtorch.evolutionary.model_training.cem import CEM, PFA, ARS, FDA, CA, PBIL

__all__ = ["DE", "GA", "CEM", "PFA", "ARS", "FDA", "CA", "PBIL"]



from typing import Any
import torch
import math
from swarmtorch.base import SwarmOptimizer


class FPA(SwarmOptimizer):
    """Flower Pollination Algorithm (FPA) optimizer for PyTorch models."""

    def __init__(
        self,
        params: Any,
        swarm_size: int = 30,
        p: float = 0.8,
        device: str = "cpu",
    ) -> None:
        super().__init__(params, swarm_size=swarm_size, device=device, p=p)
        self.p = p
        self.iteration_count = 0

    def _init_swarm(self) -> None:
        param_shape = self._get_param_shape()
        self.swarm_size = self.defaults["swarm_size"]
        self.positions = torch.rand(self.swarm_size, param_shape[0], device=self.device)
        self.fitness = torch.full((self.swarm_size,), float("inf"), device=self.device)
        self.best_position = torch.zeros(param_shape[0], device=self.device)
        self.best_fitness = torch.tensor(float("inf"), device=self.device)

    def _set_params(self, flat_params: torch.Tensor) -> None:
        idx = 0
        for group in self.param_groups:
            for p in group["params"]:
                numel = p.numel()
                p.data.copy_(flat_params[idx : idx + numel].reshape(p.shape))
                idx += numel

    def _get_params(self) -> torch.Tensor:
        params = []
        for group in self.param_groups:
            for p in group["params"]:
                params.append(p.data.flatten())
        return torch.cat(params)

    def _evaluate_fitness(
        self, particles: torch.Tensor, closure: Any = None
    ) -> torch.Tensor:
        if closure is None:
            raise ValueError("FPA requires a closure function")
        fitness = torch.zeros(particles.shape[0], device=self.device)
        for i in range(particles.shape[0]):
            self._set_params(particles[i])
            loss = closure()
            fitness[i] = loss.detach()
        return fitness

    def _levy_flight(self, dim: int) -> torch.Tensor:
        beta = 1.5
        sigma = (
            math.gamma(1 + beta)
            * math.sin(math.pi * beta / 2)
            / (math.gamma((1 + beta) / 2) * beta * 2 ** ((beta - 1) / 2))
        ) ** (1 / beta)
        u = torch.randn(dim, device=self.device) * sigma
        v = torch.randn(dim, device=self.device)
        step = u / (v.abs() ** (1 / beta))
        return step

    def _update_positions(self) -> None:
        closure = getattr(self, "_current_closure", None)
        if closure is None:
            return

        fitness = self._evaluate_fitness(self.positions, closure)
        best_idx = torch.argmin(fitness)
        if fitness[best_idx] < self.best_fitness:
            self.best_fitness = fitness[best_idx]
            self.best_position = self.positions[best_idx].clone()

        for i in range(self.swarm_size):
            if torch.rand(1, device=self.device).item() < self.p:
                self.positions[i] = self.best_position + self._levy_flight(
                    self.positions.shape[1]
                ) * (self.positions[i] - self.best_position)
            else:
                j, k = torch.randint(0, self.swarm_size, (2,)).tolist()
                self.positions[i] = self.positions[i] + torch.rand_like(
                    self.positions[i]
                ) * (self.positions[j] - self.positions[k])

        self._set_params(self.best_position)
        self.iteration_count += 1

    def step(self, closure: Any = None) -> Any:
        self._current_closure = closure
        return super().step(closure)



from swarmtorch.bio_inspired.model_training.alo import ALO
from swarmtorch.bio_inspired.model_training.bbo import BBO, MVO
from swarmtorch.bio_inspired.model_training.sine_cosine import SineCosine, MFO

__all__ = ["ALO", "BBO", "MVO", "SineCosine", "MFO"]
